do -- Section [1]: Loading process 
	FlexBar2 = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceDB-2.0", "AceModuleCore-2.0", "AceDebug-2.0")
	FlexBar2.OO = AceLibrary("AceOO-2.0");
	FlexBar2.Waterfall = AceLibrary("Waterfall-1.0");

	FlexBar2:RegisterDB("FlexBar2_DB");
	FlexBar2:RegisterDefaults('profile', {
		DataVersion = 1, -- Changed when incompatibilities occur in SavedVars during versions
		Buttons = {},
	});

	function FlexBar2:OnInitialize()
	    self:SetDebugging(true);
		self.ModuleOnEnableList = {};
		self.ButtonList = {};
		self.ButtonMixinList = {};
		self.GroupMixinList = {};
		self.ButtonActivatorList = {};
		self.ButtonDeactivatorList = {};
		self.GroupActivatorList = {};
		self.ButtonWaterfallList = {};
		self.GroupWaterfallList = {};
		self.ButtonSettingLoaderList = {};
		self.Buttons = {};
		self:Debug("Searching for modules");
	    for Name, Module in self:IterateModules() do
			self:Debug("     Found Module: " .. Name);
			if(type(Module.ButtonMixin) == "table") then
				table.insert(self.ButtonMixinList, Module.ButtonMixin);
				self:Debug("          Found ButtonMixin");
				if(type(Module.ButtonMixin.Deactivate) == "function") then
					self:Debug("               Found Activator");			
					table.insert(FlexBar2.ButtonActivatorList, Module.ButtonMixin.Activate);
				end
				if(type(Module.ButtonMixin.Deactivate) == "function") then
					self:Debug("               Found Deactivator");			
					table.insert(FlexBar2.ButtonDeactivatorList, Module.ButtonMixin.Deactivate);
				end
				if(type(Module.ButtonMixin.LoadSettings) == "function") then
					self:Debug("               Found Setting Loader");			
					table.insert(FlexBar2.ButtonSettingLoaderList, Module.ButtonMixin.LoadSettings);
				end
			end
			if(type(Module.GroupMixin) == "table") then
				self:Debug("          Found GroupMixin: ");
				self.GroupMixinList[Name] = Module.GroupMixin;
				if(type(Module.GroupMixin.Activate) == "function") then
					self:Debug("               Found Activator");
					table.insert(self.GroupActivatorList, Module.GroupMixin.Activate);
				end
			end
			if(type(Module.OnEnable) == "function") then
				self:Debug("          Found OnEnable handler");
				self.ModuleOnEnableList[Name] = Module.OnEnable;
			end
			if(type(Module.ButtonWaterfall) == "function") then
				self:Debug("          Found button waterfall");
				self.ButtonWaterfallList[Name] = Module.ButtonWaterfall;
			end
			if(type(Module.GroupWaterfall) == "function") then
				self:Debug("          Found group waterfall");
				self.GroupWaterfallList[Name] = Module.GroupWaterfall;
			end
		end
		
		self.ButtonSuperClass = self.OO.Class(unpack(self.ButtonMixinList));
			function self.ButtonSuperClass.prototype:init(Name, ...)
				FlexBar2.ButtonSuperClass.super.prototype.init(self);
				FlexBar2:Debug("          Initializing");
				if(type(Name) ~= "string") then
					FlexBar2:Debug("               Failed, invalid name");
					return false;
				end
				self.InfoTable = FlexBar2.db.profile.Buttons[Name];
				FlexBar2.Buttons[Name] = self;
				table.insert(FlexBar2.ButtonList, Name);
				for _, Func in ipairs(FlexBar2.ButtonActivatorList) do
					Func(self, ...);
				end
			end
			function self.ButtonSuperClass.prototype:LoadSettings()
				for _, Func in ipairs(FlexBar2.ButtonSettingLoaderList) do
					Func(self);
				end
			end
		self.GroupSuperClass = self.OO.Class(unpack(self.GroupMixinList));
			function self.GroupSuperClass.prototype:init(...)
				FlexBar2:Debug("          Initializing");
				FlexBar2.GroupSuperClass.super.prototype.init(self);
				for _, Func in ipairs(FlexBar2.GroupActivatorList) do
					Func(self, ...);
				end
			end
	end
	function FlexBar2:OnEnable()
		self:SetupSlashCommands();
		self:SetupWaterfall();
		self:Debug("Generating buttons");
		for Name, InfoTable in pairs(self.db.profile.Buttons) do
			if(InfoTable.Generate == true) then
				self:Debug("     Generating " .. Name);
				self:Generate(Name);
			end
		end
		for _, Func in ipairs(self.ModuleOnEnableList) do
			Func(self);
		end
	end
	function FlexBar2:OnDisable()
	    
	end

	function FlexBar2:OnModuleCreated(Name, ref)
		
	end
end

do -- Section [2]: Configuration: GUI & slashcommands 
	-- Slashcommand handler: thanks th Mairelon for the inspiration on the API (though most code was written myself except for ToTable)
	FlexBar2.SlashCommands = {};
	local SlashCommands = FlexBar2.SlashCommands;
	SlashCommands.CommandList = {};
	function SlashCommands.Dispatch(Command)
		if(type(Command) ~= "string") then FlexBar2:Print("Error: Command error"); return false; end
		local _, _, Function, Arguments = Command:find("^(%w+)[ ]*(.*)");
		if(type(Function) ~= "string" or Function == "") then FlexBar2:Print("Error: No subcommand specified"); return false; end
		if(Function == "help") then SlashCommands:Help(Arguments); end
		if(type(SlashCommands.CommandList[Function][1]) == "function") then
			local Params = SlashCommands:ToTable(Arguments or "");
			if(SlashCommands:Verify(Function, Params, true) == false) then return false; end
			SlashCommands.CommandList[Function][1](Params);
		else
			SlashCommands:Help();
		end
	end
	
	FlexBar2:RegisterChatCommand({"/fb", "/flexbar", "/fb2", "/flexbar2"},
	{
	    type = 'text',
	    name = "FlexBar2 Slashcommands",
	    desc = "Flexbar2 slashcommand handler",
	    usage = "<command>",
	    get = false,
	    set = SlashCommands.Dispatch,
	});
	
	function SlashCommands:Help(Group)
		FlexBar2:Print("Available subcommands: ");
		for Command, InfoTable in pairs(SlashCommands.CommandList) do
			if((type(Group) == "string" and Group == InfoTable.Group) or type(Group) ~= "string") then
				if(InfoTable.Help) then
					FlexBar2:Print(InfoTable.Help);
				else
					FlexBar2:Print("/FlexBar2 " .. Command .. ": no help available");
				end
			end
		end
	end

	function SlashCommands:ToTable(msg) -- nearly unmodified code from Mairelon's utility class used in FB1
		if msg == nil then return {} end
		-- pattern to return name=
		local pattern = "([%a_][%w_]*)="
		local index = 1
		local params = {}
		local firsti, lasti, capture = string.find(msg,pattern)
		-- while we have a name=, process the info after it.
		while capture do
			local varname=string.lower(capture)
			index = index+lasti
			firsti, lasti, capture = string.find(string.sub(msg, index),pattern)
			if not firsti then firsti=string.len(msg) else firsti=firsti-2 end
			local str = string.sub(msg,index, index+firsti)
			if(string.find(str, "[ ]+$")) then -- code to strip out spaces at the end
				str = string.sub(str, 1, (string.find(str, "()[ ]+$")) - 1);
			end
			-- str contain the rest of the equality ( varname = str )
			params[varname] = str;
		end
		return params;
	end

	function SlashCommands:Verify(Function, Params, Output)
		if(type(SlashCommands.CommandList[Function]) ~= "table"
			or type(SlashCommands.CommandList[Function][1]) ~= "function"
			or type(SlashCommands.CommandList[Function][4]) ~= "table") then
			if(Output) then FlexBar2:Print("Slashcommand not set up correctly, cancelling"); end
			return false;
		end
		for Param, Info in pairs(SlashCommands.CommandList[Function][4]) do
			if(type(Params[Param]) ~= "string" and Info.Required == true) then
				if(Output) then FlexBar2:Print("Required parameter [" .. Param .. "] not set, cancelling"); end
				return false;
			end
			if(type(Info.AllowedValues) == "table") then
				if(not FlexBar2:ValueIsInTable(Info.AllowedValues, Params[Param])) then
					if(Output) then FlexBar2:Print("[" .. Params[Param] .. "] is not in the list of allowed values for " .. Param); end
					return false;
				end
			end
			if(type(Info.DefaultValue) == "string" and type(Params[Param]) ~= "string") then
				Params[Param] = Info.DefaultValue;
			end
			if(type(Info.VerifyFunc) == "function" and Info.VerifyFunc(Params[Param]) == false) then
				if(Output) then FlexBar2:Print("Verification failed for [" .. Param .. "], Invalid value [" .. Params[Param] .. "]"); end
				return false;
			end
		end
	end

	--[[
		Command: The name of the command we are adding
		Function: The command that this will be passed to
		Group: What group the function is in (for the help box)
		Help: function usage which will be displayed in the /FlexBar2 help <command> message
	--]]
	function SlashCommands:AddCommand(Command, Function, Group, Help)
		if(type(Command) ~= "string" or type(Function) ~= "function" or SlashCommands.CommandList[Command] == "table") then
			return false;
		else
			Command = Command:lower();
		end
		if(type(Group) ~= "string") then Group = "main"; end
		SlashCommands.CommandList[Command] = {Function, Group, Help, {}};
	end
	
	--[[
		Command: The name of the command we are adding parameters to
		Param: The name of the parameter we are adding
		Required: wether to error out if the parameter is not set
		Types: what type the parameter can be (string or table [ranges are seen as tables])
		AllowedValues: what values are allowed for this parameter (example: {"true", "false"})
		Default: what default value the parameter takes if it is ommited
	--]]
	function SlashCommands:AddParam(Command, Param, Required, AllowedValues, DefaultValue, VerifyFunc)
		if(type(Command) ~= "string"
			or type(Param) ~= "string"
			or type(SlashCommands.CommandList[Command][4][Param]) == "table"
			or type(SlashCommands.CommandList[Command]) ~= "table") then
				return false;
		end
		SlashCommands.CommandList[Command][4][Param] =
		{
			Required = Required,
			AllowedValues = AllowedValues,
			DefaultValue = DefaultValue,
			VerifyFunc = VerifyFunc,
		};
	end
	
	local function SlashCommand_Create(Params)
		for _, ButtonName in ipairs(FlexBar2:ToButtonTable(Params.button)) do
			FlexBar2:Create(ButtonName);
		end
	end
	
	local function SlashCommand_Delete(Params)
		for _, ButtonName in ipairs(FlexBar2:ToButtonTable(Params.button)) do
			FlexBar2:Delete(ButtonName);
		end
	end
	
	function FlexBar2:SetupSlashCommands()
		SlashCommands:AddCommand("gui", function() FlexBar2.Waterfall:Open("FlexBar2"); end, "main", "/FlexBar2 gui");
		SlashCommands:AddCommand("create", SlashCommand_Create, "main", "/FlexBar2 generate button=<Buttons>");
		SlashCommands:AddParam("create", "button", true, nil, nil, function(Button) return not FlexBar2:ButtonsExist(Button); end);
		SlashCommands:AddCommand("delete", SlashCommand_Delete, "main", "/FlexBar2 delete button=<Buttons>");
			SlashCommands:AddParam("delete", "button", true, nil, nil, function(Button) return FlexBar2:ButtonsExist(Button); end);
	end
	
	function FlexBar2:SetupWaterfall() -- Called in OnEnable
		local Waterfall = self.Waterfall;
		self:Debug("Setting up Waterfall");
		local WaterfallTree = 
		{
			{
				text = "General Settings",
				id = "gnrlset",
				isOpen = true,
			},
			{
				text = "Button Control",
				id = "buttonctrl",
				isOpen = false,
			},
			{
				text = "Group Control",
				id = "groupctrl",
				isOpen = false,
			},
		};
		local buttonctrl_CreateTemp, buttonctrl_EditSave, buttonctrl_EditTemp;
		local function BuildWaterfallTree(Id)
			if not Id then return ""; end
			if(Id == "gnrlset") then
			elseif(Id == "buttonctrl") then
				if(not buttonctrl_EditSave or type(FlexBar2.Buttons[buttonctrl_EditSave]) ~= "table") then
					if(type(FlexBar2.Buttons[buttonctrl_EditSave]) ~= "table" and type(buttonctrl_EditSave) == "string") then
						FlexBar2:Print("Non-existant button");
						buttonctrl_EditSave = nil;
					end
					FlexBar2:Debug("Waterfall: ButtonCtrl main");
					-- Create Button --
					Waterfall:AddControl(
					"type", "label",
					"text", "Create Button: ");
					
					Waterfall:AddControl(
					"type","textbox",
					"width",200,
					"fullRefresh", true,
					"noNewLine", true,
					"setFunc", function() FlexBar2:Create(buttonctrl_CreateTemp); buttonctrl_EditSave = buttonctrl_CreateTemp; end,
					"getFunc", function() return ""; end,
					"changedFunc", function(ButtonName) buttonctrl_CreateTemp = ButtonName; end);
					
					Waterfall:AddControl(
					"type","button",
					"text","Create",
					"fullRefresh", true,
					"noNewLine",true,
					"width",80,
					"causeRefresh",true,
					"execFunc", function() FlexBar2:Create(buttonctrl_CreateTemp); buttonctrl_EditSave = buttonctrl_CreateTemp; end);
					
					-- Edit Button --
					Waterfall:AddControl(
					"type", "label",
					"text", "Edit Button: ");
					
					Waterfall:AddControl(
					"type","textbox",
					"width",200,
					"noNewLine", true,
					"fullRefresh", true,
					"setFunc", function(ButtonName) buttonctrl_EditSave = ButtonName end,
					"changedFunc", function(ButtonName) buttonctrl_EditTemp = ButtonName end);
					
					Waterfall:AddControl(
					"type","button",
					"text","Edit",
					"noNewLine",true,
					"fullRefresh", true,
					"width",80,
					"execFunc", function() buttonctrl_EditSave = buttonctrl_EditTemp end);
					return "Basic Button Control";
				else
					FlexBar2:Debug("Waterfall: Buttonctrl edit");
					for Module, Data in pairs(FlexBar2.ButtonWaterfallList) do
						if(type(Data) == "function") then
							Data("edit", buttonctrl_EditSave, FlexBar2:ToButtonTable(buttonctrl_EditSave));
						end
					end
					-- Delete Button --
					Waterfall:AddControl(
					"type", "button",
					"text","Delete Button",
					"fullRefresh", true,
					"width", 100,
					"execFunc", function() FlexBar2:Delete(buttonctrl_EditSave); buttonctrl_EditSave = nil; end);
					
					-- Cancel Button --
					Waterfall:AddControl(
					"type", "button",
					"text","Cancel changes",
					"fullRefresh", true,
					"noNewLine", true,
					"width", 130,
					"execFunc", function()
						buttonctrl_EditSave = nil;
						end,
					"execArg1");
				end

					-- Save Button --
					Waterfall:AddControl(
					"type", "button",
					"text","Save changes",
					"fullRefresh", true,
					"noNewLine", true,
					"width", 130,
					"execFunc", function()
						
						for Module, Data in pairs(FlexBar2.ButtonWaterfallList) do
							if(type(Data) == "function") then
								Data("save", buttonctrl_EditSave, FlexBar2:ToButtonTable(buttonctrl_EditSave));
							end
						end
						local ButtonTable = FlexBar2:ToButtonTable(buttonctrl_EditSave);
						for _, ButtonName in ipairs(ButtonTable) do
							FlexBar2.Buttons[ButtonName]:LoadSettings();
						end
						buttonctrl_EditSave = nil;
					end);

					return "Editing button(s) settings: " .. buttonctrl_EditSave;
			elseif(Id == "groupctrl") then
				
			end
		end
		Waterfall:Register('FlexBar2',
		'title', 'FlexBar2 Config',
		'width', 800,
		'height', 800,
		'tree', WaterfallTree,
		'children', BuildWaterfallTree);
	end
end

do -- Section [3]: Button control 
	function FlexBar2:Create(Name)
		if(type(FlexBar2.db.profile.Buttons[Name]) ~= "table") then
			FlexBar2.db.profile.Buttons[Name] = { Generate = true};
			FlexBar2:Generate(Name);
			return true;
		else
			FlexBar2:Print("Button already exists");
			return false;
		end
	end


	function FlexBar2:Delete(Name)
		if(type(FlexBar2.db.profile.Buttons[Name]) == "table") then
			FlexBar2.db.profile.Buttons[Name] = nil;
			local Button = FlexBar2.Buttons[Name]
			for _, Func in ipairs(FlexBar2.ButtonDeactivatorList) do
				Func(Button);
			end
			FlexBar2.Buttons[Name] = nil;
			return true;
		else
			FlexBar2:Print("Cannot delete non-existant button");
			return false;
		end
		
	end

	function FlexBar2:Generate(Name)
		self:Debug(Name);
		if(self.Buttons[Name] ~= nil) then
			self:Debug("          Error: Button already exists, aborting");
			return false;
		end
		self.ButtonSuperClass:new(Name);
		return true;
	end
end


do -- Section [4]: Utility
	function FlexBar2:GetSetting(ButtonTable, Setting)
		if(type(ButtonTable) ~= "table" or type(Setting) ~= "string") then return nil; end
		return FlexBar2.db.profile.Buttons[ButtonTable[1]][Setting];
	end

	function FlexBar2:ApplySetting(ButtonTable, Setting, Value)
		if(type(ButtonTable) ~= "table" or type(Setting) ~= "string") then return nil; end
		for _, ButtonName in ipairs(ButtonTable) do
			FlexBar2.db.profile.Buttons[ButtonName][Setting] = Value;
		end
	end
	
	function FlexBar2:ToButtonTable(ButtonRange)
		if(type(ButtonRange) == "string") then
			if(ButtonRange:find("%d+-%d+")) then
				local ButtonTable = {};
				local _, _, First, Last = ButtonRange:find("^(%d+)-(%d+)$");
				for ButtonNumber = First, Last do
					table.insert(ButtonTable, tostring(ButtonNumber));
				end
			elseif(ButtonRange:find("[%w_];[%w_]")) then
				ButtonTable = { string.split(";", ButtonRange) };
			else
				ButtonTable = { ButtonRange };
			end
			for _, ButtonName in ipairs(ButtonTable) do
				if(not ButtonName:find("^[%w_]+$")) then
					return false;
				end
			end
			return ButtonTable;
		elseif(type(ButtonRange) == "table") then
			return ButtonRange;
		else
			return false;
		end
		
	end
	
	function FlexBar2:ButtonsExist(Button)
		if(type(Button) == "string") then
			Button = FlexBar2:ToButtonTable(Button);
		end
		if(type(Button) == "table") then
			for _, ButtonName in ipairs(Button) do
				if(FlexBar2.Buttons[ButtonName] == nil) then return false; end
			end
			return true;
		else
			return false;
		end
	end
	
	function FlexBar2:ValueIsInTable(Table, Value)
		local foundValue = table.foreachi(Table, function(i,v)
			return v == Value and v or nil;
		end);
		return FoundValue == Value;
	end
end
